"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyzeVueComponentFromTemplate = void 0;
const type_analyzer_1 = require("@previewjs/type-analyzer");
const path_1 = __importDefault(require("path"));
const typescript_1 = __importDefault(require("typescript"));
function analyzeVueComponentFromTemplate(typescriptAnalyzer, getTypeAnalyzer, filePath) {
    // This virtual file exists thanks to transformReader().
    const tsFilePath = `${filePath}.ts`;
    const name = path_1.default.basename(filePath, path_1.default.extname(filePath));
    const program = typescriptAnalyzer.analyze([tsFilePath]);
    const typeAnalyzer = getTypeAnalyzer(program);
    const sourceFile = typeAnalyzer.sourceFile(tsFilePath);
    for (const statement of sourceFile.statements) {
        const definedProps = extractDefinePropsFromStatement(typeAnalyzer, statement);
        if (definedProps) {
            return {
                name,
                propsType: definedProps.type,
                providedArgs: new Set(),
                types: definedProps.collected,
            };
        }
        if (typescript_1.default.isTypeAliasDeclaration(statement) &&
            statement.name.text === "PJS_Props") {
            const type = typeAnalyzer.checker.getTypeAtLocation(statement);
            const defineComponentProps = typeAnalyzer.resolveType(type);
            return {
                name,
                propsType: defineComponentProps.type,
                providedArgs: new Set(),
                types: defineComponentProps.collected,
            };
        }
    }
    return {
        name,
        propsType: type_analyzer_1.UNKNOWN_TYPE,
        providedArgs: new Set(),
        types: {},
    };
}
exports.analyzeVueComponentFromTemplate = analyzeVueComponentFromTemplate;
function extractDefinePropsFromStatement(typeAnalyzer, statement) {
    if (typescript_1.default.isExpressionStatement(statement)) {
        // This may be a statement such as `defineProps()`.
        const definedProps = extractDefinePropsFromExpression(typeAnalyzer, statement.expression);
        if (definedProps) {
            return definedProps;
        }
    }
    if (typescript_1.default.isVariableStatement(statement)) {
        for (const variableDeclaration of statement.declarationList.declarations) {
            if (!variableDeclaration.initializer) {
                continue;
            }
            const definedProps = extractDefinePropsFromExpression(typeAnalyzer, variableDeclaration.initializer);
            if (definedProps) {
                return definedProps;
            }
        }
    }
    return null;
}
function extractDefinePropsFromExpression(typeAnalyzer, expression) {
    if (typescript_1.default.isBinaryExpression(expression) &&
        expression.operatorToken.kind === typescript_1.default.SyntaxKind.EqualsToken) {
        // This may be an assignment such as `props = defineProps()`.
        return extractDefinePropsFromExpression(typeAnalyzer, expression.right);
    }
    if (!typescript_1.default.isCallExpression(expression) ||
        !typescript_1.default.isIdentifier(expression.expression)) {
        return null;
    }
    const functionName = expression.expression.text;
    if (functionName === "defineProps") {
        const signature = typeAnalyzer.checker.getResolvedSignature(expression);
        if (signature) {
            return typeAnalyzer.resolveType(signature.getReturnType());
        }
    }
    if (functionName === "withDefaults") {
        const [firstArgument, secondArgument] = expression.arguments;
        if (!firstArgument || !secondArgument) {
            return null;
        }
        const definePropsType = extractDefinePropsFromExpression(typeAnalyzer, firstArgument);
        if (!definePropsType) {
            return null;
        }
        const defaultsType = typeAnalyzer.resolveType(typeAnalyzer.checker.getTypeAtLocation(secondArgument));
        if (defaultsType.type.kind !== "object" ||
            definePropsType.type.kind !== "object") {
            // Unsure what to do here, ignore defaults.
            return definePropsType;
        }
        const fieldsWithDefaultValue = new Set(Object.keys(defaultsType.type.fields));
        return {
            type: (0, type_analyzer_1.objectType)(Object.fromEntries(Object.entries(definePropsType.type.fields).map(([fieldKey, fieldType]) => [
                fieldKey,
                fieldsWithDefaultValue.has(fieldKey)
                    ? (0, type_analyzer_1.optionalType)(fieldType)
                    : fieldType,
            ]))),
            collected: definePropsType.collected,
        };
    }
    return null;
}
